/*****************************************************************************/
/*
                                   QandA.c

(Perhaps) another (lame) attempt to frustrate SPAMbots.

Question-and-answer challenge before form submission.  Obviously intended to
tap into knowledge shared with the submitter, or some high-level interpretation
of the question.  For example, "What's our favourite operating system?", "What
does the D in DEC stand for?", "What was the Ken ..sen family name?", etc. 
This idea is simple, has no external  dependency (like reCAPTCHA), and while
serving a real purpose could still be fun and engaging for some sites (not that
I know about such things).

Configured in VWCMS.CONF using the [qanda] directive, one question and answer
per line.  Alternate answers can be supplied separated by a vertical bar.  For
example, the above questions would be configured as

  [qanda] What's our favourite operating system?=VMS|OpenVMS
  What does the D in DEC stand for?=Digital
  What was the Ken ..sen family name?=Olsen
  The 'V' in VMS stands for?=Virtual
  What was the original hardware platform for VMS?=VAX|VAX-11

Responses are tested case-insensitive.  Though intended to apply site-wide a
qanda configuration can be made page-specific and therefore support a
different set of questions-and-answers for different submissions within the one
site.

  [if-CGI-path_info] /site/page/
     [qanda] question=answer
  [endif]


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
06-AUG-2011  MGD  QandaPresent() if not found in content look in template
23-JUL-2011  MGD  initial
*/
/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <types.h>
#include <unistd.h>

/* VMS related header files */
#include "starlet.h"
#include "ssdef.h"
#include "stsdef.h"

/* application header file */
#include "cgilib.h"
#include "config.h"
#include "content.h"
#include "data.h"
#include "form.h"
#include "other.h"
#include "template.h"
#include "qanda.h"
#include "vwcms.h"

#define FI_LI "QANDA", __LINE__

#define QANDA_MAX 100
#define QANDA_MIN   5

/* global storage */

BOOL  QandaAfter,
      QandaBefore;

static char  QandaClass [64],
             QandaStyle [256],
             QandaThese [64];

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern char  SoftwareId[],
             Utility[];

extern char  *TemplateTextPtr;

extern CONTENT_DATA  *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Implement the $$QANDA$$ directive.
*/

void QandaDirective (char* SubsArg)

{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QandaDirective() |%s|\n", SubsArg);

   if (!SubsArg)
   {
      /* initialise */
      QandaAfter = QandaBefore = FALSE;
      QandaClass[0] = QandaThese[0] = QandaStyle[0] = '\0';
      return;
   }

   if (*(cptr = SubsArg) == '$') cptr++;
   while (*cptr)
   {
      if (strsame (cptr, "class=", 6))
         cptr = TemplateParseParam (cptr+6, QandaClass, sizeof(QandaClass));
      else
      if (strsame (cptr, "these=", 6))
         cptr = TemplateParseParam (cptr+6, QandaThese, sizeof(QandaThese));
      else
      if (strsame (cptr, "style=", 6))
      {
         strcpy (QandaStyle, " style=\"");
         cptr = TemplateParseParam (cptr+6, QandaStyle, sizeof(QandaStyle)-10);
         strcat (QandaStyle, "\"");
      }
      else
      if (strsame (cptr, "after=submit", 12))
      {
         QandaAfter = TRUE;
         QandaBefore = FALSE;
         cptr = TemplateParseSkipParam (cptr+12, NULL);
      }
      else
      if (strsame (cptr, "before=submit", 13))
      {
         QandaAfter = FALSE;
         QandaBefore = TRUE;
         cptr = TemplateParseSkipParam (cptr+13, NULL);
      }
      else
      if (strsame (cptr, "submit", 6))
      {
         QandaAfter = FALSE;
         QandaBefore = TRUE;
         cptr = TemplateParseSkipParam (cptr+6, NULL);
      }
      else
      {
         TemplateParamError ("QANDA", SubsArg, cptr);
         return;
      }
   }
   if (!QandaAfter && !QandaBefore) QandaBefore = TRUE;
}

/*****************************************************************************/
/*
Add the question-and-answer challenge to the form.
*/

void QandaChallenge ()

{
   static int  LastQanda = 0;
   static unsigned long  RandomNumber,
                         RandomFiller;

   int  status,
        QandaCount = 0;
   char  *cptr, *sptr,
         *ContentPtr;
   char  *QandaArray [QANDA_MAX];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QandaChallenge()\n");

   if (QandaThese[0])
   {
      cptr = MakeFileName (QandaThese, ".LIS");
      status = DataReadFileIntoMemory (cptr, &ContentPtr, NULL, TRUE);
      if (VMSnok (status))
      {
         TemplateAdvise ("EQANDA !AZ (%X!8XL)", SysGetMsg(status), status);
         return;
      }
      cptr = ContentPtr;
   }
   else
   if (!(cptr = VWcmsConfig.QandaCollectionPtr))
      cptr = "";

   while (*cptr)
   {
      if (QandaCount >= QANDA_MAX) break;
      QandaArray[QandaCount++] = cptr;
      while (*cptr && *cptr != '\n') cptr++;
      while (*cptr && *cptr == '\n') cptr++;
   }
   if (QandaCount < QANDA_MIN)
   {
      TemplateAdvise ("EQANDA configuration error");
      return;
   }

   /* select one at random (but never the same two consecutively) */
   if (!(RandomNumber & 0xff)) sys$gettim (&RandomNumber);
   while ((RandomNumber % QandaCount) == LastQanda)
      RandomNumber = RandomNumber * 69069 + 1;
   cptr = QandaArray [LastQanda = (RandomNumber % QandaCount)];

   for (sptr = cptr; *sptr && *sptr != '='; sptr++);
   if (!*sptr)
   {
      TemplateAdvise ("EQANDA configuration error");
      return;
   }
   *sptr = '\0';

   fprintf (stdout,
"<p><fieldset class=\"%s\"%s>\
<legend>&nbsp;%s&nbsp;</legend>\
<input name=\"qanda_field_one\" type=\"hidden\" value=\"%s\">\
%s&nbsp;&nbsp;\
<input name=\"qanda_field_two\" type=\"text\" size=\"15\" value=\"\" \
autocomplete=\"off\">\
</fieldset></p>\n",
            QandaClass[0] ? QandaClass : "vwcms_qanda_challenge",
            QandaStyle,
            CONFIG_LANG_PARSE(LangQanda,1),
            HtmlEscape(cptr),
            HtmlEscape(cptr));

   *sptr = '=';
}

/*****************************************************************************/
/*
Get the question-and-answer form fields (question asked and response supplied)
and then go searching for the question in the list of challenge questions. 
When found compare the supplied answer and return NULL is matching.  Otherwise
return a string indicating the error.
*/

char* QandaVerify ()

{
   int  status;
   char  *cptr, *sptr, *zptr,
         *AnswerField,
         *ContentPtr,
         *QuestionField;
   char  **taptr;
   char  *TextArray[] = { ContentSectionPtr->TextPtr, TemplateTextPtr, NULL };
   char  AnswerBuffer [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QandaVerify()\n");

   QuestionField = GetCgiVarNull ("FORM_QANDA_FIELD_ONE");
   AnswerField = GetCgiVarNull ("FORM_QANDA_FIELD_TWO");
   if (!QuestionField || !QuestionField[0] || !AnswerField)
      return ("bugcheck!");

   if (!AnswerField[0]) return (CONFIG_LANG_PARSE(LangQanda,2));

   /* look for the directive and check if it specifies a file */
   for (taptr = TextArray; *taptr; taptr++)
   {
      cptr = TemplateParseFind (*taptr, "$$QANDA$");
      if (*cptr) cptr+= 8;
      while (*cptr && *(USHORTPTR)cptr != '$$')
      {
         cptr = TemplateParseFindParam (cptr, "these=");
         if (strsame (cptr, "these=", 6))
         {
            TemplateParseParam (cptr+6, QandaThese, sizeof(QandaThese));
            break;
         }
      }
   }

   if (QandaThese[0])
   {
      cptr = MakeFileName (QandaThese, ".LIS");
      status = DataReadFileIntoMemory (cptr, &ContentPtr, NULL, TRUE);
      if (VMSnok (status))
      {
         cptr = CgiLibVeeMemCalloc (256);
         sprintf (cptr, "%s (%%%08.08X)", SysGetMsg(status), status);
         return (cptr);
      }
      cptr = ContentPtr;
   }
   else
   if (!(cptr = VWcmsConfig.QandaCollectionPtr))
      cptr = "";

   while (*cptr)
   {
      for (sptr = cptr; *sptr && *sptr != '='; sptr++);
      if (!*sptr) break;
      *sptr = '\0';
      if (!strcmp (cptr, QuestionField))
      {
         *sptr++ = '=';
         cptr = sptr;
         zptr = (sptr = AnswerBuffer) + sizeof(AnswerBuffer)-2;
         while (*cptr && *cptr != '\n' && sptr < zptr) *sptr++ = *cptr++;
         *(USHORTPTR)sptr = '\0\0';
         for (cptr = AnswerBuffer; *cptr; cptr = sptr)
         {
            for (sptr = cptr; *sptr && *sptr != '|'; sptr++);
            *sptr++ = '\0';
            if (strsame (cptr, AnswerField, -1)) return (NULL);
         }
         /* answer did not match */
         return (CONFIG_LANG_PARSE(LangQanda,3));
      }
      else
         *sptr++ = '=';
      while (*sptr && *sptr != '\n') sptr++;
      while (*sptr && *sptr == '\n') sptr++;
      cptr = sptr;
   }

   /* hmmm, the question was not found! */
   return ("Question not found!");
}

/*****************************************************************************/
/*
Returns a pointer if the section content or template contains a $$QANDA$$
directive, else NULL.
*/

char* QandaPresent ()

{
   char  *cptr;
   char  **taptr;
   char  *TextArray[] = { ContentSectionPtr->TextPtr, TemplateTextPtr, NULL };

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QandaPresent()\n");

   for (taptr = TextArray; *taptr; taptr++)
      if (*(cptr = TemplateParseFind (*taptr, "$$QANDA$")))
         return (cptr);

   return (NULL);
}

/*****************************************************************************/

